#!/usr/bin/env python

##
##Rich - 21/10/08
##
##Little test rig to MOSDEF code, throw it through cparse.py and cparse2.py
## generate the IL and then compare the results. The hope being that this will
## help quickly ID the areas cparse2 is sucking at or just getting plain wrong.

##
import sys, time, difflib, getopt, hotshot, os

if "." not in sys.path: sys.path.append(".")

try:
    from engine import CanvasConfig as config
    config["cparse_version"]="3"
except ImportError:
    ##For Standalone MOSDEF
    print "Standalone MOSDEF"
    config={"cparse_version":"3"}

import cparse, cparse2, mosdef


class timer:
    def __init__(self):
        ##Results stored as ID : [start, stop, delta]
        self.results={}

    def start(self, id):
        ##start the clock
        self.results[id]=[time.time()]

    def stop(self, id):
        ##Stop the clock + calculate the delta
        self.results[id].append(time.time())
        self.results[id].append( self.results[id][1] - self.results[id][0] )

    def get(self, id):
        ##return the tiings related to the given ID
        try:
            return self.results[id]
        except:
            return ["Unknown ID [%s]"%(id)]


def show_IL(code_1, code_2):
    """
    Print out the complete IL listing that was generated
    in two coloumns
    """
    code_1=code_1.splitlines(1)
    code_2=code_2.splitlines(1)
    ##Remove any lines that are JUST \n
    try:
        while 1:
            code_1.remove('\n')
    except ValueError:
        pass #all '\n's gone

    try:
        while 1:
            code_2.remove('\n')
    except ValueError:
        pass #all '\n's gone

    #print code_1[:10]
    #print "\n"
    #print code_2[:10]

    ##Work out column width
    width=0
    for l in code_1:
        if len(l) > width:
            width = len(l)	    
    width+=2

    ##Work out which code listing is longer for loop below
    if len(code_1) > len(code_2):
        code_len=len(code_1)
    else:
        code_len=len(code_2)

    print "IL Listing: "
    for line_no in range(0,code_len):
        try:
            print " %s%s|\t"%(code_1[line_no].replace("\n","")," "*(width-len(code_1[line_no])-1)),
        except IndexError:
            print " %s|\t"%(" "*(width)),

        try:
            print "%s"%(code_2[line_no].replace("\n",""))
        except IndexError:
            print ""


def show_metrics(run1, run2,col_width=30):

    listing1=run1[1].splitlines(1)
    listing2=run2[1].splitlines(1)
    try:
        while 1:
            listing1.remove('\n')
    except ValueError:
        pass #all '\n's gone
    try:
        while 1:
            listing2.remove('\n')
    except ValueError:
        pass #all '\n's gone      

    print "\nIL Metrics:",
    print "%scparse1\t\tcparse2"%(" "*(col_width-12))
    print "Number of lines of IL:",
    print "%s%d\t\t%d"%(" "*(col_width-23),len(listing1), len(listing2))
    print "Mean time to generate IL:",
    print "%s%f secs\t%f secs\tdelta: %f secs"%(" "*(col_width-26),run1[0], run2[0], (run1[0]-run2[0]) )
    print "Percentage time differences:",
    print "%s%.2f %%\t\t%.2f %%\t\t"%(" "*(col_width-29),(run1[0]/run2[0])*100, (run2[0]/run1[0])*100 )

    print "\n"

def show_diff(run1, run2):
    ##Print out a diff of the generated IL between cparse1 and cparse2
    listing1=run1.splitlines(1)
    listing2=run2.splitlines(1)
    try:
        while 1:
            listing1.remove('\n')
    except ValueError:
        pass #all '\n's gone
    try:
        while 1:
            listing2.remove('\n')
    except ValueError:
        pass #all '\n's gone    

    #d=difflib.context_diff(run1.splitlines(1), run2.splitlines(1))
    d=difflib.context_diff(listing1, listing2)
    diff_str=''.join(d)

    print "\n\nIL Code diff:\n"
    if len(diff_str) == 0:
        print "**Generated: IDENTICAL IL - woohoo!**\n\n"
        None
    else:
        split_diff(diff_str)


def split_diff(context_diff_str, f1_mark='*', f2_mark='-'):
    ##Take a context diff string and format it for side by side viewing

    mark1="%s "%(f1_mark*3)
    mark2="%s "%(f2_mark*3)
    division="***************"
    start=0

    #print context_diff_str
    chunk={}
    chunk_num=0
    col_width=0
    context_diff_str=context_diff_str.splitlines(1)

    ##Split into the separate chunks of the diff structure
    for line in context_diff_str:

        ##Little skip for false positives as LABEL's will always have different ID's
#	if "LABEL" in line:
#	    continue

        ##Auto-adjust column width for pretty printing to accom the longest string in the set
        if len(line) > col_width:
            col_width=len(line)+2
        if mark1 in line:
            ##Start of chunk
            chunk_num+=1
            chunk[chunk_num]=[]
            chunk[chunk_num].append(line.replace('\n',''))
        elif mark2 in line:
            ##End of chunk
            chunk_num+=1
            chunk[chunk_num]=[]
            chunk[chunk_num].append(line.replace('\n',''))
        elif division in line:
            pass
        else:
            ##Content of chunk
            chunk[chunk_num].append(line.replace('\n',''))

    print "cparse1%s|"%(" "*(col_width-7)),
    print "\tcparse2"
    ##Iterate through and display them side by side
    for index in range(1,chunk_num, 2):
        line_no=0
        ##Skip empty chunks that just have markers
        if len(chunk[index]) == 1:
            continue

        while 1:
            try:
                print "%s%s|"%(chunk[index][line_no]," "*(col_width-len(chunk[index][line_no]))),
                print "\t%s"%(chunk[index+1][line_no])
                line_no+=1
            except IndexError:
                print ""
                break

def get_cparse1(filename, t, var_dict):

    mosdef.USE_CPARSE2=False
    config["cparse_version"]="1"

    t.start('cparse1')
    list = {}
    list['linppc']    = {'proc':"ppc", 'os':"Linux"}
    list['linx86'] = {'proc':"intel", 'os':"Linux"}

    data=open(filename).read()    
    # opening remoteresolver module
    rr = mosdef.getremoteresolver(list['linx86']['os'], list['linx86']['proc'])

    vars=var_dict

    defines={}
    data = mosdef.preprocess(data, vars, defines, rr, delim = "//")

    tokens=cparse.scan(data)
    #print tokens
    tree=cparse.parse(tokens)

    generator=cparse.ilgenerate(tree,vars,remoteresolver=rr)
    cp1_IL=generator.get()
    t.stop('cparse1')
    timings=t.get('cparse1')
    return [timings, cp1_IL]

def get_cparse2(filename, t, var_dict):

    mosdef.USE_CPARSE2=True
    config["cparse_version"]="2"

    t.start('cparse2')    
    data=file(filename,"r").read()

    ##instantiate a parser, yaccer and parse the data we say - for texting this deafults to Win32 & X86
    #OS="Win32"
    OS="Linux"
    PROC="X86"
    cp2_IL=cparse2.getCparser(target_os=OS, target_proc=PROC, vars=var_dict, code=data)

    t.stop('cparse2')

    timings=t.get('cparse2')
    return [timings, cp2_IL]


def usage():
    print "Usage: %s -m <MOSDEF source> -v <MOSDEF variables> [-d (dump entire IL file)]"%(sys.argv[0])
    sys.exit(1)

def parse_opts():
    """
    Grab options from CLI
    """
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hdpv:m:", ["help", "dump", "vars","mosdef","profile"])
    except getopt.GetoptError, err:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit(2)

    dump = False
    profile=False
    mdf=None
    var_file=None
    c_vars={}
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        elif o in ("-d", "--dump"):
            dump=True
        elif o in ("-v", "--vars"):
            ##A python dictionary of vars to pass into the parser
            ## useful for #import stuff    
            var_file=a
        elif o in ("-m", "--mosdef"):
            ##The mosdef c file that is being assembled
            mdf=a
        elif o in ("-p", "--profile"):
            ##For profiling the code to find slow spots use a default arg set
            #mdf="MOSDEF/test_win32.c"
            #var_file="MOSDEF/test_win32_2.vars"
            profile=True
        else:
            assert False, "Unknown option"

    if not mdf:
        print "You must specify a MOSDEF source file to assemble"
        usage()

    ##If we haven't explicitly declared a var file use a default name one same as the .c but .var
    if not var_file:
        try:
            mdf=args[0]
            mdf=mdf.split('.')
            print "No explicit variable file defined, falling back to: %s.vars"%(mdf[0])
            var_file="%s.vars"%(mdf[0])
        
            

            ##Read in and exec the var file    
            try:
                vf=open(var_file,'r')
            except IOError, err:
                print "Unable to find variable definition file: %s"%(var_file)
                sys.exit(-1)
            while 1:
                line=vf.readline()
                if not line:
                    break
                var_list=line.split('=')
                exec("c_vars['%s']=%s"%(var_list[0],var_list[1]))
            print "Variables defined: %s"%(c_vars)    
            
        except IndexError:
            ##No vars defined
            print "No C vars defined"
            c_vars={}
            
            
    ##The only arg we support atm is that of the test.c file
    return mdf,c_vars,dump, profile

if __name__ == "__main__":

    mosdef_file,c_vars,dump, profile=parse_opts()

    timer_obj=timer()
    ##How many times shall we run the code gen to get a mean?
    num_runs=1
    r1_times=[]
    r2_times=[]

    for x in range(0,num_runs):

        if not profile:
            ret_1=get_cparse1(mosdef_file, timer_obj, c_vars)
            r1_times.append(ret_1[0][2])        
            ret_2=get_cparse2(mosdef_file, timer_obj, c_vars)
            r2_times.append(ret_2[0][2])
            #ret_1=ret_2
            #r1_imes=r2_times
        else:
            print "cparse 1 profile:"
            pr=hotshot.Profile("cparse1_stats")
            pr.runcall(get_cparse1, mosdef_file, timer_obj, c_vars)
            pr.close()
            from hotshot import stats
            s=stats.load("cparse1_stats")
            s.sort_stats("time").print_stats(25)

            print "="*80
            print "cparse 2 profile:"
            pr=hotshot.Profile("cparse2_stats")
            pr.runcall(get_cparse2, mosdef_file, timer_obj, c_vars)
            pr.close()
            s=stats.load("cparse2_stats")
            s.sort_stats("time").print_stats(25)
            sys.exit(0)
    print "------------------"
    ##Get mean time and write this to our final results lists
    ##Add in mode and range?
    tmp=0
    for t in r1_times:
        tmp+=t
    ret_1[0]=tmp/num_runs
    tmp=0
    for t in r2_times:
        tmp+=t
    ret_2[0]=tmp/num_runs


    ##Display the results
    show_diff(ret_1[1], ret_2[1])
    show_metrics(ret_1, ret_2)
    if dump:
        show_IL(ret_1[1], ret_2[1])
